/* GNSS Logging
 *  Main code to control de Open drifter.
 
 * Created = 23/07/2018 JFuentes
 * Modified = 11/11/2022 JFuentes
 * 
 * Microcontroller board: SAMD21-MINI
 
 */

/************ Serial 2 ************/
//Serial 2 send lines to the comunication devices used, LoRa or GSM.
#include <Adafruit_SleepyDog.h>
#include <Arduino.h>   // required before wiring_private.h
#include "wiring_private.h" // pinPeripheral() function

int smartNumber = 5; //Numer of the drifter
 
Uart Serial2 (&sercom1, 11, 10, SERCOM_RX_PAD_0, UART_TX_PAD_2);

void SERCOM1_Handler()
{
  Serial2.IrqHandler();
}

/************ Libraries ************/
//SD card
#include <SPI.h> //For I2C conection
#include <SD.h> //Specific for SD card

//Digital sensor libraries
#include <Wire.h>

#include <Adafruit_Sensor.h>
#include <Adafruit_BNO055.h>
#include <utility/imumaths.h>

#include <Adafruit_GFX.h>    // Core graphics library
#include <Adafruit_SSD1306.h>

/************ Variables ************/

int serialPlot = 0; //For debugging, equal to 1 activates the serial monitoring

//IMU BNO055:

imu::Quaternion quat;
imu::Vector<3> acc;
imu::Vector<3> magnet;
imu::Vector<3> gyros;

uint8_t syste, gyro, accel, mag = 0;

Adafruit_BNO055 bno = Adafruit_BNO055(55,0x29);

//#define BNO055_SAMPLERATE_DELAY_MS (10)//Set the delay between fresh samples
//Adafruit_BNO055 bno = Adafruit_BNO055();

//Digital pins:

int NewFile = 4; //Pin to control the creation of a new data logging file
const int chipSelect = 9; //Where the SD card is conected, specific from shield 

//GNSS:

#define DATETIME_FORMAT "%04d.%02d.%02d %02d:%02d:%02d"
#define DATETIME_LENGTH 20

char datetime[DATETIME_LENGTH];

//GNSS Message Structure:
//The GNSS must be configured before to send NAV_PAT messages in the highest sampling rate: U-center software from u-blox.

struct NAV_PVT {
  
    unsigned char cls;
    unsigned char id;
    unsigned short len;
    unsigned long iTOW;          // GPS time of week of the navigation epoch (ms)
 
    unsigned short yea;         // Year (UTC)
    unsigned char mont;         // Month, range 1..12 (UTC)
    unsigned char da;           // Day of month, range 1..31 (UTC)
    unsigned char hou;          // Hour of day, range 0..23 (UTC)
    unsigned char minut;        // Minute of hour, range 0..59 (UTC)
    unsigned char secon;        // Seconds of minute, range 0..60 (UTC)
    char valid;                  // Validity Flags (see graphic below)
    unsigned long tAcc;          // Time accuracy estimate (UTC) (ns)
    long nano;                   // Fraction of second, range -1e9 .. 1e9 (UTC) (ns)
    unsigned char fixType;       // GNSSfix Type, range 0..5
    char flags;                  // Fix Status Flags
    char flags2;                  // Fix Status Flags
    unsigned char numSV;         // Number of satellites used in Nav Solution
 
    long lon;                    // Longitude (deg)
    long lat;                    // Latitude (deg)
    long height;                 // Height above Ellipsoid (mm)
    long hMSL;                   // Height above mean sea level (mm)
    unsigned long hAcc;          // Horizontal Accuracy Estimate (mm)
    unsigned long vAcc;          // Vertical Accuracy Estimate (mm)
 
    long velN;                   // NED north velocity (mm/s)
    long velE;                   // NED east velocity (mm/s)
    long velD;                   // NED down velocity (mm/s)
    long gSpeed;                 // Ground Speed (2-D) (mm/s)
    long heading;                // Heading of motion 2-D (deg)
    
    unsigned long sAcc;          // Speed Accuracy Estimate
    unsigned long headingAcc;    // Heading Accuracy Estimate
    unsigned short pDOP;         // Position dilution of precision

    unsigned char reserved[6] = {0,0,0,0,0,0};
    long headVeh;
    short magDeg;
    unsigned short magAcc;
};

NAV_PVT pvt;

//Data logging files:

String FileName = ""; //Variable to save the name of our logging file  
File dataFile; //To save the file that we open 

//Logging and sampling variables:

unsigned long TIME_SYNC = 10000; //How many seconds do we need to wait to load collected data to SD card (it makes the adquisition faster)
unsigned long TIME_FLUSH = 0; //Time used as reference in sync, it will change every sync cycle

//Display:

#define OLED_RESET 4
Adafruit_SSD1306 display(OLED_RESET);


/************ Setup ************/

void setup() {
  
  Watchdog.enable(5000); //Resets the Microcontroller board  if there is a failure.

  // Put your setup code here, to run once:
  
  SerialUSB.begin(9600); //Define The serial connection
  Serial1.begin(9600); //GNSS

  //If you want to wait until serial is connected uncomment: 
//  if (serialPlot==1)
//  {
//    while (!SerialUSB) {
//      ; // Wait for serial port to connect. Needed for native USB port only
//    }
//  }
  
  // Serial 2 for comunications GPR and LoRa:
  
  Serial2.begin(9600);
  Watchdog.reset();
  // Assign pins 10 & 11 SERCOM functionality
  pinPeripheral(10, PIO_SERCOM);
  pinPeripheral(11, PIO_SERCOM);

  // Display:
  
  display.begin(SSD1306_SWITCHCAPVCC, 0x3C);  // initialize with the I2C addr 0x3C (for the 128x32)
  // Clear the buffer.
  display.clearDisplay();
  display.setTextSize(1);
  display.setTextColor(WHITE);
  display.println("Open Particle V1.0");
  display.display();
  
  Watchdog.reset();

  // SD Card configuration:
  
  SerialUSB.print("Initializing SD card...");
  
  // see if the card is present and can be initialized:
  if (!SD.begin(chipSelect)) {
    SerialUSB.println("Card failed, or not present");
    // don't do anything more:
    return;
  }
  
  Watchdog.reset();
  
  SerialUSB.println("...card initialized.");

  //File name selection:

  int i = 0; //Controls the number of files
  int Exist = 1; //Controls the existance of files 
  
  for(i=1; Exist==1 & i<100;i++) //99 Files allowed
  {
    FileName = "FILE"+ String(i) + ".TXT";
    Exist = SD.exists(FileName);
  }
  
  //Detect if it has been reached the maximum number of files allowed
  if (Exist==1)
  {
    SerialUSB.println("Maximum number of files reached.");
    display.clearDisplay();
    display.setCursor(0,0);
    //  display.setTextSize(1);
    //  display.setTextColor(WHITE);
    
    display.println("Max number of file reached!");
    display.display();

    
    while(1) {};
  }
  else
  {
    dataFile = SD.open(FileName, FILE_WRITE);
    
    //Detects if there was an error opening the file
    if( ! dataFile ) 
    {   
      //SerialUSB.println("Error opening the file.");
      display.clearDisplay();
      display.setCursor(0,0);
      //  display.setTextSize(1);
      //  display.setTextColor(WHITE);
      
      display.println("Error opening the file.");
      display.display();

      while(1) {};
    }
    else
    {
      SerialUSB.println(FileName);
      
      // Header of the file:     
      dataFile.println("Arduino (ms)\t GPS (ms)\tTime UTC (ms)\tNano UTC (ns)\tlat (1e-7 deg)\tlon (1e-7 deg)\tHeight above elipsoide (mm)\tHorizontal accuracy est. (mm)\tGround speed (mm/s)\tv_N (mm/s)\tv_E (mm/s)\tv_D (mm/s)\tSpeed accuracy (mm/s)\tNumber of satellites\tHeading (1e-5 deg)\tQw\tQy\tQx\tQz\tAccx(m/s^2)\tAccy(m/s^2)\tAccz(m/s^2)\tMagx(uT)\tMagy(uT)\tMagz(uT)\tGyrox(rad/s)\tGyroy(rad/s)\tGyroz(rad/s)\tCalGyro\tCalAccel\tCalMag");
    }
  }
  Watchdog.reset();
  
  /************ Sensors initialization ************/
  
  // Initialise the IMU: 
  
  if(!bno.begin())
  {
    /* There was a problem detecting the BNO055 ... check your connections */
    SerialUSB.print("Ooops, no BNO055 detected ... Check your wiring or I2C ADDR!");
    while(1);
  }
  
  //bno.setExtCrystalUse(true);
  
  //New file bottom:
  
  pinMode(NewFile, INPUT);
  
  Watchdog.reset();
}

/*Loop*/

void loop() {
  
  TIME_FLUSH = millis(); // SD sync time
  
  int senseValue = digitalRead(NewFile); //Detects new file request
  
  int sensorCount = 0; //To control long push of the button
  
  /***** Sync Control *****/
  
  while (millis()<TIME_FLUSH+TIME_SYNC) //Until TIME_SYNC dont save the data in the file
  {
    
    //Put your main code after this: 

    millisTime = millis();
    
    if (processGPS()) 
    {   
      senseValue = digitalRead(NewFile);
    
      if (senseValue == HIGH)
      {
        sensorCount++;
      }
      else
      {
        sensorCount = 0;
      } 
    
      if(sensorCount==2) //New file
      {
        
        SerialUSB.println("New File");
        dataFile.close();

        int i = 0; //Controls the number of files
        int Exist = 1; //Controls the existance of files 
        
        for(i=1; Exist==1 & i<100;i++) //99 Files allowed
        {
          FileName = "FILE"+ String(i) + ".TXT";
          Exist = SD.exists(FileName);
        }
        
        //Detect if it has been reached the maximum number of files allowed
        if (Exist==1)
        {
          //Display maximum number of files reached
          display.clearDisplay();
          display.setCursor(0,0);
          //  display.setTextSize(1);
          //  display.setTextColor(WHITE);
  
          display.println("Max number of file reached!");
          display.display();
          
          while(1) {};
        }
        else
        {
          if (serialPlot==1)
          {
            SerialUSB.println(FileName);
          }
          
          dataFile = SD.open(FileName, FILE_WRITE);
          
          //Detects if there was an error opening the file
          if( ! dataFile ) 
          {   
            display.clearDisplay();
            display.setCursor(0,0);
           
            display.println("Error opening the file.");
            display.display();
            
            while(1) {};
          }
          else
          {
            SerialUSB.println(FileName);

            display.clearDisplay();
            display.setCursor(0,0);
            display.print(FileName);
            display.println(" created!");
            display.display();
            delay(5000);
            
            //Header     
            dataFile.println("Arduino (ms)\t GPS (ms)\tTime UTC (ms)\tNano UTC (ns)\tlat (1e-7 deg)\tlon (1e-7 deg)\tHeight above elipsoide (mm)\tHorizontal accuracy est. (mm)\tGround speed (mm/s)\tv_N (mm/s)\tv_E (mm/s)\tv_D (mm/s)\tSpeed accuracy (mm/s)\tNumber of satellites\tHeading (1e-5 deg)\tQw\tQy\tQx\tQz\tAccx(m/s^2)\tAccy(m/s^2)\tAccz(m/s^2)\tMagx(uT)\tMagy(uT)\tMagz(uT)\tGyrox(rad/s)\tGyroy(rad/s)\tGyroz(rad/s)\tCalGyro\tCalAccel\tCalMag");
          }
        }
        while(Serial1.available()){Serial1.read();} // empty GNSS Serial...
      }

      /***** IMU *****/
      
      // Quaternion data
      //imu::Quaternion 
      quat = bno.getQuat();
      // Accel data
      //imu::Vector<3> 
      acc = bno.getVector(Adafruit_BNO055::VECTOR_LINEARACCEL);
      //Magnetometer
      magnet = bno.getVector(Adafruit_BNO055::VECTOR_MAGNETOMETER);
      //Gyroscope
      gyros = bno.getVector(Adafruit_BNO055::VECTOR_GYROSCOPE);
      //Cal
      bno.getCalibration(&syste, &gyro, &accel, &mag);
      
      /***** GNSS *****/
      snprintf(datetime, DATETIME_LENGTH, DATETIME_FORMAT, pvt.yea, pvt.mont, pvt.da, pvt.hou, pvt.minut, pvt.secon);
      
      if (serialPlot==1) //Plots via serial if activated
      {
        //Arduino time
        SerialUSB.print(millis());
        SerialUSB.print("\t");
        
        //GPS
        SerialUSB.print(pvt.iTOW);
        SerialUSB.print("\t");
        SerialUSB.print(datetime);
        SerialUSB.print("\t");
        SerialUSB.print(pvt.nano);
        SerialUSB.print("\t");
        SerialUSB.print(pvt.lat);//divide by 10000000.0f
        SerialUSB.print("\t");
        SerialUSB.print(pvt.lon); //divide by 10000000.0f
        SerialUSB.print("\t");
        SerialUSB.print(pvt.height);
        SerialUSB.print("\t");
//        SerialUSB.print(pvt.hMSL);
//        SerialUSB.print("\t");
        SerialUSB.print(pvt.hAcc);
        SerialUSB.print("\t"); 
//        SerialUSB.print(pvt.vAcc);    
//        SerialUSB.print("\t");
        SerialUSB.print(pvt.gSpeed);
        SerialUSB.print("\t");      
        SerialUSB.print(pvt.velN);
        SerialUSB.print("\t");
        SerialUSB.print(pvt.velE);
        SerialUSB.print("\t");
        SerialUSB.print(pvt.velD);
        SerialUSB.print("\t");
        SerialUSB.print(pvt.sAcc);
        SerialUSB.print("\t");
        SerialUSB.print(pvt.numSV);
        SerialUSB.print("\t"); 
        SerialUSB.print(pvt.heading);
        SerialUSB.print("\t");     
        
        //IMU
        SerialUSB.print(quat.w(), 4);
        SerialUSB.print("\t");
        SerialUSB.print(quat.y(), 4);    
        SerialUSB.print("\t");    
        SerialUSB.print(quat.x(), 4);    
        SerialUSB.print("\t");    
        SerialUSB.print(quat.z(), 4);
        SerialUSB.print("\t");
        SerialUSB.print(acc.x());    
        SerialUSB.print("\t");    
        SerialUSB.print(acc.y());
        SerialUSB.print("\t");
        SerialUSB.print(acc.z());
        SerialUSB.print("\t");
        SerialUSB.print(magnet.x());    
        SerialUSB.print("\t");    
        SerialUSB.print(magnet.y());
        SerialUSB.print("\t");
        SerialUSB.print(magnet.z());
        SerialUSB.print("\t");
        SerialUSB.print(gyros.x());    
        SerialUSB.print("\t");    
        SerialUSB.print(gyros.y());
        SerialUSB.print("\t");
        SerialUSB.print(gyros.z());

        SerialUSB.println();      
      }
      
      //Arduino time
      dataFile.print(millis());
      dataFile.print("\t");
      
      //GPS
      dataFile.print(pvt.iTOW);
      dataFile.print("\t");
      dataFile.print(datetime);
      dataFile.print("\t");
      dataFile.print(pvt.nano);
      dataFile.print("\t");
      dataFile.print(pvt.lat);//10000000.0f
      dataFile.print("\t");
      dataFile.print(pvt.lon);//10000000.0f
      dataFile.print("\t");      
      dataFile.print(pvt.height);
      dataFile.print("\t");       
//      dataFile.print(pvt.hMSL);
//      dataFile.print("\t"); 
      dataFile.print(pvt.hAcc);
      dataFile.print("\t");
//      dataFile.print(pvt.vAcc);
//      dataFile.print("\t");
      dataFile.print(pvt.gSpeed);
      dataFile.print("\t");
      dataFile.print(pvt.velN);
      dataFile.print("\t");
      dataFile.print(pvt.velD);
      dataFile.print("\t");
      dataFile.print(pvt.velE);
      dataFile.print("\t");
      
      dataFile.print(pvt.sAcc);
      dataFile.print("\t");
      dataFile.print(pvt.numSV);
      dataFile.print("\t");
      dataFile.print(pvt.heading);
      dataFile.print("\t");  
      
      //IMU
      dataFile.print(quat.w(), 4);
      dataFile.print("\t");
      dataFile.print(quat.y(), 4);
      dataFile.print("\t");
      dataFile.print(quat.x(), 4);
      dataFile.print("\t");
      dataFile.print(quat.z(), 4);
      dataFile.print("\t");
      dataFile.print(acc.x());
      dataFile.print("\t");
      dataFile.print(acc.y());
      dataFile.print("\t");
      dataFile.print(acc.z());     
      dataFile.print("\t");
      
      dataFile.print(magnet.x());
      dataFile.print("\t");
      dataFile.print(magnet.y());
      dataFile.print("\t");
      dataFile.print(magnet.z());     
      dataFile.print("\t");

      dataFile.print(gyros.x());
      dataFile.print("\t");
      dataFile.print(gyros.y());
      dataFile.print("\t");
      dataFile.print(gyros.z());     
      dataFile.print("\t");
            
      dataFile.print(gyro, 4);
      dataFile.print("\t");
      dataFile.print(accel, 4);
      dataFile.print("\t");
      dataFile.print(mag, 4);

      dataFile.println();
     
      while(Serial1.available()){Serial1.read();} // empty GNSS Serial...
      
      //if comunication microcontroller asks for a line send it:
      if (Serial2.available() > 0) 
      {
        char c = Serial2.read();
        if (c == 'l') {
          Serial2.print(smartNumber);
          Serial2.print(" ");
          Serial2.print(pvt.lat);//10000000.0f
          Serial2.print(" ");
          Serial2.print(pvt.lon);//10000000.0f
          Serial2.print(" ");
          Serial2.print(pvt.gSpeed);//10000000.0f
          Serial2.print("f");
          }
      }
      Watchdog.reset();
    }
    
    Serial1.flush();
  }

  //After TIME_SYNC the data is saved to the SD and variables updated in the screen:
  
  dataFile.flush(); // Data saved to the SD
  
  Watchdog.reset();

  display.clearDisplay();
  display.setCursor(0,0);
  
  display.print("hAcc = ");display.print(pvt.hAcc/1000.00);display.println(" m");
  
  display.print("Speed = ");display.print(pvt.gSpeed/1000.00);display.println(" m/s ");
  display.print("Sat = ");display.print(pvt.numSV);display.println("");
  
  display.print("Cal = ");display.print("G:");display.print(gyro);display.print("/");display.print("A:");display.print(accel);display.print("/");display.print("M:");display.print(mag);display.println("");
  display.display();

}
